"""Build c_resolve cython extension."""


def build(setup_kwargs):
    """Needed for the poetry building interface."""

    try:
        from Cython.Build import cythonize
    except ImportError:
        pass
    else:
        # use cythonize to build the extensions
        modules = ["src/buvar/di/c_di.pyx", "src/buvar/components/c_components.pyx"]

        extensions = cythonize(modules)
        setup_kwargs.update({"ext_modules": extensions, "include_dirs": []})



import functools
import logging
import logging.config
from os import getpid
import sys
import typing

import attr
import structlog
from structlog.processors import JSONRenderer

from . import util


def stringify_dict_keys(obj):
    if isinstance(obj, list):
        obj = [stringify_dict_keys(item) for item in obj]
    elif isinstance(obj, tuple):
        obj = tuple(stringify_dict_keys(item) for item in obj)
    elif isinstance(obj, dict):
        obj = {str(key): stringify_dict_keys(value) for key, value in obj.items()}
    return obj


try:
    import orjson

    json_dumps_inner = functools.partial(
        orjson.dumps, option=orjson.OPT_NAIVE_UTC | orjson.OPT_NON_STR_KEYS
    )
    json_dumps = lambda *arg, **kwargs: json_dumps_inner(*arg, **kwargs).decode("utf-8")
except ImportError:
    import json

    json_dumps = json.dumps


LOGGING_LEVEL_NAMES = list(
    map(
        logging.getLevelName,
        sorted(
            (
                logging.NOTSET,
                logging.DEBUG,
                logging.INFO,
                logging.WARN,
                logging.ERROR,
                logging.CRITICAL,
            )
        ),
    )
)
DEFAULT_LOGGING_LEVEL = logging.getLevelName(logging.WARNING)
timestamper = structlog.processors.TimeStamper(fmt="ISO", utc=True)


@attr.s(auto_attribs=True, kw_only=True)
class LogConfig:
    tty: bool = sys.stdout.isatty()
    level: typing.Union[int, str] = logging.DEBUG
    user_config: typing.Optional[typing.Dict] = None
    capture_warnings: bool = True
    redirect_print: bool = False
    json_renderer: JSONRenderer = JSONRenderer(
        serializer=lambda obj, **kwargs: json_dumps(stringify_dict_keys(obj), **kwargs)
    )

    @property
    def processors(self):
        processors = [
            structlog.stdlib.add_log_level,
            structlog.stdlib.add_logger_name,
            structlog.stdlib.PositionalArgumentsFormatter(),
            timestamper,
            add_os_pid,
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
        ]
        return processors

    @property
    def pre_chain(self):
        pre_chain = [
            # Add the log level and a timestamp to the event_dict if the log entry
            # is not from structlog.
            structlog.stdlib.add_logger_name,
            structlog.stdlib.add_log_level,
            structlog.processors.format_exc_info,
            timestamper,
            add_os_pid,
            # ExtractLogExtra(
            #     "spec",
            #     "url",
            #     "mimetype",
            #     "has_body",
            #     "swagger_yaml",
            #     "method",
            #     "path",
            #     "operation_id",
            #     "data",
            # ),
        ]
        return pre_chain

    def setup(self):
        level = (
            logging.getLevelName(self.level.upper())
            if isinstance(self.level, str)
            else self.level
        )

        renderer = structlog.dev.ConsoleRenderer() if self.tty else self.json_renderer
        config = {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "structured": {
                    "()": structlog.stdlib.ProcessorFormatter,
                    "processor": renderer,
                    "foreign_pre_chain": self.pre_chain,
                }
            },
            "handlers": {
                "default": {"class": "logging.StreamHandler", "formatter": "structured"}
            },
            "loggers": {
                "": {"handlers": ["default"], "level": level, "propagate": True}
            },
        }
        if self.user_config:
            util.merge_dict(self.user_config, dest=config)
        logging.config.dictConfig(config)

        logging.captureWarnings(self.capture_warnings)

        structlog.configure(
            processors=self.processors,
            context_class=structlog.threadlocal.wrap_dict(dict),
            logger_factory=structlog.stdlib.LoggerFactory(),
            wrapper_class=structlog.stdlib.BoundLogger,
            cache_logger_on_first_use=True,
        )

        if self.redirect_print:
            # redirect stdio print
            print_log = structlog.get_logger("print")
            sys.stderr = StdioToLog(print_log)
            sys.stdout = StdioToLog(print_log)

        # log uncaught exceptions
        sys.excepthook = uncaught_exception


def setup_logging(**kwargs):
    log_config = LogConfig(**kwargs)
    log_config.setup()


class ExtractLogExtra:  # noqa: R0903

    """Extract log record attributes to structlog event_dict."""

    def __init__(self, *attrs):
        self.attrs = attrs

    def __call__(self, logger, method_name, event_dict):
        """
        Add the logger name to the event dict.
        """
        record = event_dict.get("_record")
        for attr_name in self.attrs:
            if hasattr(record, attr_name):
                attr = getattr(record, attr_name)
                event_dict[attr_name] = attr
        return event_dict


class StdioToLog:

    """Delegate sys.stdout to a logger."""

    def __init__(self, logger, log_level=logging.INFO):
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ""

    def write(self, buf):  # noqa
        for line in buf.rstrip().splitlines():
            self.logger.log(self.log_level, line.rstrip())

    def flush(self):  # noqa
        pass


def add_os_pid(logger, method_name, event_dict):  # noqa
    event_dict["pid"] = getpid()
    return event_dict


def uncaught_exception(ex_type, ex_value, tb):  # noqa: C0103
    log_ = structlog.get_logger("sys.excepthook")
    log_.critical(event="uncaught exception", exc_info=(ex_type, ex_value, tb))



import dataclasses
import functools
import os
import sys
import typing

import attr
import cattr
import structlog

from . import di, util

logger = structlog.get_logger()


@attr.s(auto_attribs=True)
class ConfigValue:
    name: typing.Optional[str] = None
    help: typing.Optional[str] = None


class ConfigSource(dict):
    """Config dict, with loadable sections.

    >>> @attr.s(auto_attribs=True)
    ... class FooConfig:
    ...     bar: float
    >>>
    >>>
    >>> source = {'foo': {'bar': 1.23}}
    >>> config = ConfigSource(source)
    >>> foo = config.load(FooConfig, "foo")
    >>> foo
    FooConfig(bar=1.23)
    """

    __slots__ = ("env_prefix",)

    def __init__(self, *sources, env_prefix: typing.Optional[str] = None):
        super().__init__()
        util.merge_dict(*sources, dest=self)
        # config = schematize(__source, cls, env_prefix=env_prefix)

        self.env_prefix: typing.Tuple[str, ...] = (env_prefix,) if env_prefix else ()

    def merge(self, *sources):
        util.merge_dict(*sources, dest=self)

    def load(self, config_cls, name=None):
        if name is None:
            values = self
        else:
            values = self.get(name, {})

        env_config = create_env_config(
            config_cls, *(self.env_prefix + ((name,) if name else ()))
        )
        values = util.merge_dict(values, env_config)
        # merge environment

        config = relaxed_converter.structure(values, config_cls)
        return config


class ConfigError(Exception): ...


# to get typing_inspect.is_generic_type()
ConfigType = typing.TypeVar("ConfigType", bound="Config")

skip_section = type.__new__(
    type, "skip_section", (object,), {"__repr__": lambda self: self.__class__.__name__}
)()


class Config:
    __buvar_config_section__: typing.Optional[str] = skip_section
    __buvar_config_sections__: typing.Dict[str, type] = {}

    def __init_subclass__(cls, *, section: str = skip_section, **_):
        if section is skip_section:
            return
        if section in cls.__buvar_config_sections__:
            raise ConfigError(
                f"Config section `{section}` already defined!",
                cls.__buvar_config_sections__,
                cls,
            )
        cls.__buvar_config_section__ = section
        cls.__buvar_config_sections__[section] = cls

    @classmethod
    async def adapt(cls: typing.Type[ConfigType], source: ConfigSource) -> ConfigType:
        config = source.load(cls, cls.__buvar_config_section__)
        return config


def get_fields(cls):
    if dataclasses.is_dataclass(cls):
        return dataclasses.fields(cls)
    elif attr.has(cls):
        return attr.fields(cls)
    raise TypeError("Only dataclasses or attrs supported", cls)


def traverse_attrs(cls, *, target=None, get_type_hints=typing.get_type_hints):
    """Traverse a nested attrs structure, create a dictionary for each nested
    attrs class and yield all fields resp. path, type and target dictionary."""
    stack = [
        (
            target if target is not None else {},
            (),
            list(get_fields(cls)),
            get_type_hints(cls),
        )
    ]
    while stack:
        target, path, fields, hints = stack.pop()
        while fields:
            field = fields.pop()
            field_path = path + (field.name,)
            field_type = hints[field.name]
            if attr.has(field_type):
                target[field.name] = field_target = {}
                # XXX should we yield also attrs classes?
                yield field_path, field_type, target

                stack.append((target, path, fields, hints))
                target, path, fields, hints = (
                    field_target,
                    field_path,
                    list(get_fields(field_type)),
                    get_type_hints(field_type),
                )
            else:
                yield field_path, field_type, target


def create_env_config(cls, *env_prefix):
    frame = sys._getframe(1)
    get_type_hints = functools.partial(
        typing.get_type_hints, globalns=frame.f_globals, localns=frame.f_locals
    )

    env_config = {}
    for path, _, target in traverse_attrs(
        cls, target=env_config, get_type_hints=get_type_hints
    ):
        env_name = "_".join(map(lambda x: x.upper(), env_prefix + path))
        if env_name in os.environ:
            logger.debug("Overriding config by env", var=env_name)
            target[path[-1]] = os.environ[env_name]
    return env_config


# FIXME: deprecate relaxed_converter
converter = relaxed_converter = cattr.Converter()


def _env_to_bool(val, type):
    """
    Convert *val* to a bool if it's not a bool in the first place.
    """
    if isinstance(val, type):
        return val
    elif isinstance(val, str):
        val = val.strip().lower()
        if val in ("1", "true", "yes", "on"):
            return True

    return False


relaxed_converter.register_structure_hook(bool, _env_to_bool)


def generate_env_help(cls, env_prefix=""):
    """Generate a list of all environment options."""

    help = "\n".join(  # noqa: W0622
        "_".join((env_prefix,) + path if env_prefix else path).upper()
        for path, type, _ in traverse_attrs(cls)
        if not attr.has(type)
    )
    return help


async def prepare():
    di.register(Config.adapt)



import pytest


def test_forked_parent(mocker):
    import os
    from buvar import fork

    mocker.patch("os.getpid", return_value=1)
    mock_pgid = mocker.patch("os.getpgid", return_value=2)
    mock_waitid = mocker.patch("os.waitid")
    mocker.patch("os.fork", return_value=123)
    f = fork.Fork(2)
    f.run(lambda x: x)

    mock_pgid.assert_called_with(1)
    mock_waitid.assert_called_with(os.P_PGID, 2, os.WEXITED)
    assert f.children == {123}


def test_forked_child(mocker):
    from buvar import fork

    mocker.patch("os.waitid")
    mocker.patch("os.fork", return_value=0)
    f = fork.Fork(2)

    stuff_args = {}

    def stuff(a, b):
        stuff_args[a] = b

    f.run(stuff, 1, 2)
    assert stuff_args == {1: 2}


def test_forked_no_fork(mocker):
    from buvar import fork

    f = fork.Fork(1)

    stuff_args = {}

    def stuff(a, b):
        stuff_args[a] = b

    f.run(stuff, 1, 2)
    assert stuff_args == {1: 2}



class ComponentLookupError(Exception):
    pass


try:
    # gains over 100% speed up
    from .c_components import Components
except ImportError:
    from .py_components import Components  # noqa: F40



import pytest


# @pytest.mark.skipif(
#     sys.version_info < (3, 7),
#     reason="similar to https://github.com/python/typing/issues/506",
# )
@pytest.mark.asyncio
async def test_config_source_schematize(mocker):
    import typing

    import attr

    from buvar import config

    @attr.s(auto_attribs=True)
    class FooConfig:
        bar: str = "default"
        foobar: float = 9.87
        baz: bool = False

    @attr.s(auto_attribs=True)
    class BarConfig:
        bim: float
        foo: FooConfig = FooConfig()

    @attr.s(auto_attribs=True, kw_only=True)
    class BimConfig:
        bar: BarConfig
        bam: bool
        bum: int = 123
        lst: typing.List

    sources = [
        {"bar": {"bim": "123.4", "foo": {"bar": "1.23", "baz": "true"}}},
        {"foo": {"bar": "value", "foobar": 123.5, "baz": True}},
        {
            "bar": {"bim": "123.4"},
            "bim": {"bar": {"bim": 1.23}, "bam": "on", "lst": [1, 2, 3]},
        },
    ]

    mocker.patch(
        "os.environ",
        {
            "PREFIX_BAR_BIM": "0",
            "PREFIX_BAR_FOO_FOOBAR": "7.77",
            "PREFIX_BAR_FOO_BAZ": "false",
        },
    )

    cfg = config.ConfigSource(*sources, env_prefix="PREFIX")
    # foo_config = await adapters.nject(FooConfig, source=cfg)

    bim = cfg.load(BimConfig, "bim")
    bar = cfg.load(BarConfig, "bar")
    foo = cfg.load(FooConfig, "foo")
    # (FooConfig(bar="value", foobar=123.5, baz=True),)

    assert (bar, foo, bim) == (
        BarConfig(bim=0.0, foo=FooConfig(bar="1.23", foobar=7.77, baz=False)),
        FooConfig(bar="value", foobar=123.5, baz=True),
        BimConfig(bar=BarConfig(bim=1.23), bam=True, lst=[1, 2, 3]),
    )


@pytest.mark.asyncio
@pytest.mark.buvar_plugins("buvar.config")
async def test_config_generic_adapter(mocker):
    import typing

    import attr

    from buvar import config, di

    mocker.patch.dict(config.Config.__buvar_config_sections__, clear=True)

    @attr.s(auto_attribs=True)
    class FooConfig(config.Config, section="foo"):
        bar: str = "default"
        foobar: float = 9.87
        baz: bool = False

    @attr.s(auto_attribs=True)
    class BarConfig(config.Config, section="bar"):
        bim: float
        foo: FooConfig = FooConfig()

    @attr.s(auto_attribs=True, kw_only=True)
    class BimConfig(config.Config, section="bim"):
        bar: BarConfig
        bam: bool
        bum: int = 123
        lst: typing.List

    sources = [
        {"bar": {"bim": "123.4", "foo": {"bar": "1.23", "baz": "true"}}},
        {"foo": {"bar": "value", "foobar": 123.5, "baz": True}},
        {
            "bar": {"bim": "123.4"},
            "bim": {"bar": {"bim": 1.23}, "bam": "on", "lst": [1, 2, 3]},
        },
    ]

    mocker.patch(
        "os.environ",
        {
            "PREFIX_BAR_BIM": "0",
            "PREFIX_BAR_FOO_FOOBAR": "7.77",
            "PREFIX_BAR_FOO_BAZ": "false",
        },
    )

    cfg = config.ConfigSource(*sources, env_prefix="PREFIX")
    foo_config = await di.nject(FooConfig, source=cfg)
    assert foo_config == FooConfig(bar="value", foobar=123.5, baz=True)


def test_load_general_config():
    import typing

    import attr

    from buvar import config

    sources = [{"foo": "bar", "group": {"some": "value"}}]
    cfg = config.ConfigSource(*sources, env_prefix="PREFIX")

    @attr.s(auto_attribs=True)
    class GeneralVars:
        foo: str
        baz: typing.Optional[float] = None

    general = cfg.load(GeneralVars)
    assert general == GeneralVars("bar", None)


def test_config_missing():
    import attr
    from cattrs.errors import ClassValidationError

    from buvar import config

    source: dict = {"foo": {}}

    @attr.s(auto_attribs=True)
    class FooConfig:
        bar: str

    cfg = config.ConfigSource(source)
    with pytest.raises(ClassValidationError):
        cfg.load(FooConfig, "foo")


@pytest.mark.xfail
def test_generate_toml_help():
    import typing

    import attr

    from buvar import config

    @attr.s(auto_attribs=True)
    class FooConfig:
        """FooConfig.

        bim bam
        """

        string_val: str
        float_val: float = 9.87
        bool_val: bool
        int_val: int
        list_val: typing.List

    @attr.s(auto_attribs=True)
    class BarConfig:
        """BarConfig.

        bla bla
        bli bli
        """

        bim: float
        foo: FooConfig

    env_vars = {}
    config_fields = list(config.traverse_attrs(BarConfig, target=env_vars))

    assert {path for path, _, _ in config_fields} == {
        ("foo",),
        ("bim",),
        ("foo", "string_val"),
        ("foo", "float_val"),
        ("foo", "bool_val"),
        ("foo", "int_val"),
        ("foo", "list_val"),
    }

    help = config.generate_toml_help(BarConfig)
    assert (
        help.as_string()
        == """# BarConfig.
#
# bla bla
# bli bli

# bim =


[foo]
# FooConfig.
#
# bim bam

# string
# string_val =

# float
float_val = 9.87

# bool
# bool_val =

# int
# int_val =

# list
list_val = []

"""
    )  # noqa: W291


def test_nested_attrs_typing():
    import typing

    import attr

    from buvar import config

    @attr.s(auto_attribs=True)
    class Baz:
        baz: str = "foobar"

    @attr.s(auto_attribs=True)
    class Bar:
        baz: Baz

    @attr.s(auto_attribs=True)
    class Foo:
        bars: typing.List[Bar] = []

    source = config.ConfigSource(
        {"foo": {"bars": [{"baz": {"baz": "something else"}}]}}, env_prefix="TEST"
    )

    foo = source.load(Foo, "foo")
    assert foo == Foo(bars=[Bar(baz=Baz(baz="something else"))])


def test_env_config(mocker):
    import attr

    from buvar import config

    @attr.s(auto_attribs=True)
    class Baz:
        float: float

    @attr.s(auto_attribs=True)
    class Foo:
        str: str
        int: int
        baz: "Baz"

    @attr.s(auto_attribs=True)
    class Bar:
        foo: "Foo"

    mocker.patch("os.environ", {"PREFIX_FOO_STR": "abc", "PREFIX_FOO_INT": "777"})
    env_config = config.create_env_config(Bar, "PREFIX")

    assert env_config == {"foo": {"baz": {}, "str": "abc", "int": "777"}}


@pytest.mark.asyncio
@pytest.mark.buvar_plugins("buvar.config")
async def test_config_subclass_abc(mocker):
    import abc

    import attr

    from buvar import config, di

    mocker.patch.dict(config.Config.__buvar_config_sections__, clear=True)

    class GeneralConfig(config.Config, section=None): ...

    class FooBase(metaclass=abc.ABCMeta):
        @abc.abstractmethod
        def foo(self): ...

    @attr.s(auto_attribs=True)
    class FooConfig(config.Config, FooBase, section="foo"):
        bar: str

        def foo(self): ...

    assert config.skip_section not in config.Config.__buvar_config_sections__
    assert FooBase not in config.Config.__buvar_config_sections__.values()
    assert config.Config.__buvar_config_sections__["foo"] is FooConfig
    cfg = config.ConfigSource({"foo": {"bar": "abc"}}, env_prefix="PREFIX")
    foo_config = await di.nject(FooConfig, source=cfg)

    assert foo_config == FooConfig(bar="abc")


@pytest.mark.asyncio
@pytest.mark.buvar_plugins("buvar.config")
async def test_config_dataclass(mocker):
    import dataclasses as dc

    from buvar import config, di

    mocker.patch.dict(config.Config.__buvar_config_sections__, clear=True)

    @dc.dataclass
    class FooConfig(config.Config, section="foo"):
        bar: str

    cfg = config.ConfigSource({"foo": {"bar": "abc"}}, env_prefix="PREFIX")
    foo_config = await di.nject(FooConfig, source=cfg)
    assert foo_config == FooConfig(bar="abc")



import pytest


@pytest.mark.parametrize(
    "obj, result",
    [
        (None, None),
        (1, 1),
        (1.2, 1.2),
        (True, True),
        ((1, False, {1: 2}), (1, False, {"1": 2})),
        ({1: 1, 1.2: 1.2, None: None}, {"1": 1, "1.2": 1.2, "None": None}),
        ({1: {2: 3}}, {"1": {"2": 3}}),
    ],
)
def test_stringify_dict_keys(obj, result):
    from buvar.log import stringify_dict_keys

    stringified = stringify_dict_keys(obj)
    assert stringified == result


def _test_simple_structlog_json(capsys, mocker, caplog):
    from buvar import log
    import logging
    import structlog
    import json

    mocker.patch(
        "structlog.processors._make_stamper", return_value=lambda event_dict: event_dict
    )
    mocker.patch(
        "buvar.log.add_os_pid",
        side_effect=lambda logger, method_name, event_dict: event_dict,
    )

    log.setup_logging(tty=False)

    sl = structlog.get_logger("foobar")
    logging.debug("foobar: %s", 123)
    sl.info("message", foo={123: "bar"})
    captured = capsys.readouterr()
    msgs = list(map(json.loads, captured.err.strip().split("\n")))
    assert msgs == [
        {"event": "foobar: 123", "level": "debug", "logger": "root"},
        {
            "event": "message",
            "foo": {"123": "bar"},
            "level": "info",
            "logger": "foobar",
        },
    ]



import pytest


def test_run(event_loop):
    from buvar import plugin, context, ComponentLookupError

    result = plugin.stage("tests.foo_plugin", loop=event_loop)
    assert result == [{"foo": "foo"}, None, None]
    with pytest.raises(ComponentLookupError) as e:
        context.get("foo_plugin")
    assert e.value.args[1] == "foo_plugin"


def test_run_task_factory(event_loop):
    import asyncio
    from buvar import plugin, context

    async def prepare():
        async def task():
            async def _sub1():
                context.add("foo")
                await asyncio.sleep(0)
                assert "foo" == context.get(str)

            async def _sub2():
                context.add("bar")
                await asyncio.sleep(0)
                assert "bar" == context.get(str)

            await asyncio.gather(
                asyncio.create_task(_sub1()), asyncio.create_task(_sub2())
            )

        yield task()

    result = plugin.stage(prepare, loop=event_loop)


def test_run_relative_out_of_packages(event_loop):
    from buvar import plugin

    with pytest.raises(ImportError):
        plugin.stage("tests.baz_plugin", loop=event_loop)


def test_run_with_loader(event_loop, mocker):
    from buvar import plugin

    async def test_plugin(load: plugin.Loader):
        pass

    plugin.stage(test_plugin)


def test_run_load_twice(event_loop):
    from buvar import plugin

    loaded = {}

    async def test_plugin(include: plugin.Loader):
        assert not loaded
        loaded[True] = True
        await include(test_plugin)

    plugin.stage(test_plugin)


def test_run_iterable(event_loop):
    from buvar import plugin

    state = {}

    async def iter_plugin():
        async def a():
            state["a"] = True

        async def b():
            state["b"] = True

        return [a(), b()]

    plugin.stage(iter_plugin, loop=event_loop)
    assert state == {"a": True, "b": True}


def test_plugin_error(event_loop):
    from buvar import plugin

    state = {}

    async def broken_plugin(teardown: plugin.Teardown):
        async def teardown_task():
            state["teardown_task"] = True

        teardown.add(teardown_task())

        raise Exception("Plugin is broken")

    with pytest.raises(Exception) as e:
        plugin.stage(broken_plugin, loop=event_loop)
        assert e.error.args == ("Plugin is broken",)
    assert state == {"teardown_task": True}


def test_cancel_staging(event_loop):
    import asyncio
    from buvar import plugin

    state = {}

    async def server_plugin(cancel: plugin.Cancel):
        async def server():
            # shutdown
            cancel.set()
            try:
                await asyncio.Future()
            except asyncio.CancelledError:
                state["cancelled"] = True

        yield server()

    plugin.stage(server_plugin, loop=event_loop)
    assert state == {"cancelled": True}


def test_resolve_plugin_not_async(event_loop):
    from buvar import plugin

    with pytest.raises(ValueError):
        plugin.resolve_plugin_func(lambda: None)


def test_subtask(event_loop):
    from buvar import plugin

    state = {}

    async def test_plugin(teardown: plugin.Teardown):
        state["plugin"] = True

        async def teardown_task():
            state["teardown_task"] = True

        teardown.add(teardown_task())

        async def task():
            state["task"] = True

        yield task()

    plugin.stage(test_plugin, loop=event_loop)

    assert state == {"plugin": True, "task": True, "teardown_task": True}


def test_broken_task(event_loop):
    import asyncio
    from buvar import plugin

    state = {}

    async def broken_plugin(teardown: plugin.Teardown):
        fut_task = asyncio.Future()

        async def task():
            await fut_task

        async def teardown_task():
            state["teardown_task"] = True
            fut_task.cancel()

            try:
                await fut_task
            except asyncio.CancelledError:
                ...

        teardown.add(teardown_task())

        async def broken_task():
            raise Exception("Task is broken")

        yield task()
        yield broken_task()

    with pytest.raises(Exception) as e:
        plugin.stage(broken_plugin, loop=event_loop)
        assert e.error.args == ("Task is broken",)
    assert state == {"teardown_task": True}


def test_stage_components():
    import asyncio
    from buvar import plugin, components, context

    async def test_plugin(load: plugin.Loader):
        assert context.get(str) == "foo"
        context.add("bar", name="bar")

        async def task1():
            assert context.get(str, name="bar") == "bar"
            asyncio.sleep(0.02)
            context.add("task1", name="bar")
            asyncio.sleep(0.01)
            assert context.get(str, name="bar") == "task1"

        async def task2():
            assert context.get(str, name="bar") == "bar"
            asyncio.sleep(0.01)
            context.add("task2", name="bar")
            asyncio.sleep(0.02)
            assert context.get(str, name="bar") == "task2"

        yield task1()
        yield task2()

    cmps = components.Components()
    cmps.add("foo")

    plugin.stage(test_plugin, components=cmps)



__version__ = "0.43.22"
__version_info__ = tuple(__version__.split("."))


from . import context, di  # noqa: F401
from .components import ComponentLookupError, Components  # noqa: F401
from .config import ConfigSource  # noqa: F401
from .plugin import Cancel, Teardown, run, stage



"""
Fixtures and markers for testing.

`buvar_plugins` mark

.. code-block::

    @pytest.mark.buvar_plugins("buvar.config")
"""

import pytest

PLUGINS_MARK = "buvar_plugins"


def pytest_configure(config):
    config.addinivalue_line(
        "markers", f"{PLUGINS_MARK}(*plugins): run the test in buvar plugin context"
    )


def pytest_runtest_setup(item):
    # use fixtures: buvar_plugin_context
    if (
        PLUGINS_MARK in item.keywords
        and buvar_plugin_context.__name__ not in item.fixturenames
    ):
        item.fixturenames.append(buvar_plugin_context.__name__)


@pytest.fixture
def buvar_config_source():
    from buvar import config

    config_source = config.ConfigSource()

    return config_source


@pytest.fixture
def buvar_context(buvar_config_source):
    from buvar import components

    context = components.Components()
    context.add(buvar_config_source)

    return context


@pytest.fixture
def buvar_stage(event_loop, buvar_context):
    from buvar import plugin

    stage = plugin.Stage(loop=event_loop, components=buvar_context)
    return stage


@pytest.fixture
def buvar_load(request, buvar_stage):
    """Load plugins marked via `pytest.mark.buvar_plugins`."""
    # get plugins from mark
    plugins = next(
        (
            mark.args
            for mark in request.node.iter_markers()
            if mark.name == PLUGINS_MARK
        ),
        (),
    )

    try:
        # stage 1: bootstrap plugins
        if plugins:
            buvar_stage.load(*plugins)

        # stage 2: tasks
        yield buvar_stage.loader
    except Exception:
        raise

    finally:
        # stage 3: teardown
        buvar_stage.run_teardown()


@pytest.fixture
def buvar_plugin_context(buvar_stage, buvar_load):
    """The shared context while plugins are prepared."""
    return buvar_stage.context


@pytest.hookimpl(hookwrapper=True, trylast=True)
def pytest_pyfunc_call(pyfuncitem):
    """Wrap marked tests in buvar plugin context."""
    if PLUGINS_MARK in pyfuncitem.keywords:
        plugin_context = pyfuncitem.funcargs[buvar_plugin_context.__name__]
        pyfuncitem.obj = wrap_in_buvar_plugin_context(plugin_context, pyfuncitem.obj)

    yield


def wrap_in_buvar_plugin_context(context, func):
    """Enable test function to run in plugin context."""
    import contextvars
    import functools

    from buvar.context import buvar_context

    ctx = contextvars.copy_context()

    @functools.wraps(func)
    def inner(**kwargs):
        def wrapper():
            buvar_context.set(context)
            return func(**kwargs)

        return ctx.run(wrapper)

    return inner


@pytest.fixture
def buvar_adapters_setup_contextvars():
    """Set and reset adapters.
    This fixture should be used, if you need to reset adapter registration for a test.
    """
    from buvar import di

    token = di.buvar_adapters.set(di.Adapters())
    yield
    di.buvar_adapters.reset(token)



import pytest


@pytest.mark.asyncio
async def test_tasks_with_context_child():
    import asyncio
    from buvar import context

    context.add("foo")
    context.add(123)

    async def task_1():
        await asyncio.sleep(0.02)
        context.add("bar")
        await asyncio.sleep(0.01)
        assert context.get(str) == "bar"
        assert context.get(int) == 123

    async def task_2():
        await asyncio.sleep(0.01)
        context.add("baz")
        await asyncio.sleep(0.02)
        assert context.get(str) == "baz"
        assert context.get(int) == 123

    with context.child():
        fut_1 = asyncio.create_task(task_1())

    with context.child():
        fut_2 = asyncio.create_task(task_2())

    await fut_1
    await fut_2

    assert context.get(str) == "foo"


@pytest.mark.asyncio
async def test_tasks_context_auto_stacking(event_loop):
    import asyncio
    from buvar import context

    context.set_task_factory(loop=event_loop)
    try:
        context.add("foo")
        context.add(123)

        async def task_1():
            await asyncio.sleep(0.02)
            context.add("bar")
            await asyncio.sleep(0.01)
            assert context.get(str) == "bar"
            assert context.get(int) == 123

        async def task_2():
            await asyncio.sleep(0.01)
            context.add("baz")
            await asyncio.sleep(0.02)
            assert context.get(str) == "baz"
            assert context.get(int) == 123

        fut_1 = asyncio.create_task(task_1())
        fut_2 = asyncio.create_task(task_2())

        await fut_1
        await fut_2
    finally:
        event_loop.get_task_factory().reset()
        assert event_loop.get_task_factory() is None


def test_global_context_child():
    from buvar import context

    context.add("foo")
    context.add(123)
    with context.child():
        context.add("bar")
        assert context.get(str) == "bar"
        assert context.get(int) == 123
    assert context.get(str) == "foo"



import functools
import importlib
import inspect
import sys
import types
import typing as t


def methdispatch(func):
    # https://stackoverflow.com/questions/24601722/how-can-i-use-functools-singledispatch-with-instance-methods
    dispatcher = functools.singledispatch(func)

    def wrapper(*args, **kw):
        return dispatcher.dispatch(args[1].__class__)(*args, **kw)

    wrapper.register = dispatcher.register
    functools.update_wrapper(wrapper, func)
    return wrapper


class adict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self


class cached:
    def __init__(self, wrapped):
        self.wrapped = wrapped
        functools.update_wrapper(self, wrapped)

    def __get__(self, inst, cls=None):
        if inst is None:
            return self
        val = self.wrapped(inst)
        setattr(inst, self.wrapped.__name__, val)
        return val


def merge_dict(*sources, dest=None):
    """Merge `sources` into `dest`.

    `dest` is altered in place.
    """
    if dest is None:
        dest = {}
    for source in sources:
        for key, value in source.items():
            if isinstance(value, dict):
                # get node or create one
                node = dest.setdefault(key, {})
                merge_dict(value, dest=node)
            else:
                dest[key] = value
    return dest


def resolve_dotted_name(
    name: str, *, caller: t.Union[types.FrameType, int] = 0
) -> t.Union[types.ModuleType, t.Callable]:
    """Use pkg_resources style dotted name to resolve a name."""
    # skip resolving for module and coroutine
    if inspect.ismodule(name) or inspect.isroutine(name):
        return name

    # relative import
    if name.startswith("."):
        # find coller package
        frame = (
            caller if isinstance(caller, types.FrameType) else sys._getframe(1 + caller)
        )
        caller_package = frame.f_globals["__package__"]
    else:
        caller_package = None

    part = ":"
    module_name, _, attr_name = name.partition(part)

    if part in attr_name:
        raise ValueError(f"Invalid name: {name}", name)

    try:
        resolved = importlib.import_module(module_name, caller_package)
    except ValueError as ex:
        raise ImportError(*ex.args)

    if attr_name:
        resolved = getattr(resolved, attr_name)

    return resolved


def fqdn(obj) -> str:
    return f"{obj.__module__}.{obj.__qualname__}"



"""A simple background job manager.

Every background job gets added to the set and removed when done.
You may use a :py:obj:`asyncio.Semaphore` to deal with concurrency."""
import asyncio
import sys

import structlog

from buvar import context, plugin, util


class Jobs(set):
    def add(self, job, *, sync=None):
        frame = sys._getframe(1)
        module_name = frame.f_globals["__name__"]

        fut_job = asyncio.ensure_future(self.job(sync, job))

        super().add(fut_job)
        fut_job.add_done_callback(self.remove)
        sl = structlog.get_logger()
        sl.info("Added background job", job=job, module=module_name, count=len(self))

        return fut_job

    @util.methdispatch
    def job(self, sync, job):
        return job

    @job.register
    def semaphore_job(self, semaphore: asyncio.Semaphore, job):
        async def semaphore_job():
            async with semaphore:
                return await job

        return semaphore_job()

    def remove(self, fut):
        try:
            _ = fut.result()
        except Exception as ex:
            sl = structlog.get_logger()
            sl.error("Background job failed", exc_info=ex)
        finally:
            super().remove(fut)

    def cancel(self):
        for job in self:
            job.cancel()

    def __await__(self):
        return asyncio.gather(*self, return_exceptions=True).__await__()

    async def shutdown(self):
        self.cancel()
        await self


async def prepare(teardown: plugin.Teardown):
    jobs = context.add(Jobs())
    teardown.add(jobs.shutdown())



import pytest


@pytest.fixture
def adapters():
    from buvar.di import Adapters

    return Adapters()


def test_adapters_register_generic_factory(adapters):
    import typing
    from buvar import di

    FooType = typing.TypeVar("FooType", bound="Foo")

    class Foo:
        @classmethod
        async def generic_adapt(cls: typing.Type[FooType]) -> FooType:
            return cls()

    adapters.register(Foo.generic_adapt)
    assert (
        next(iter(adapters[di.ClassmethodAdapter][Foo])).implementation
        == Foo.generic_adapt
    )


def test_adapters_register_class(adapters):
    from buvar import di

    class Bar:
        def __init__(self):
            ...

    adapters.register(Bar)
    assert next(iter(adapters[di.GenericAdapter][Bar])).implementation == Bar


def test_adapters_register_classmethod(adapters):
    from buvar import di

    class Bar:
        @classmethod
        def adapt(cls) -> "Bar":
            return cls()

    adapters.register(Bar.adapt)
    assert next(iter(adapters[di.ClassmethodAdapter][Bar])).implementation == Bar.adapt


def test_adapters_register_func(adapters):
    from buvar import di

    class Bar:
        ...

    def foo_adapter() -> Bar:
        return Bar()

    adapters.register(foo_adapter)
    assert next(iter(adapters[di.GenericAdapter][Bar])).implementation == foo_adapter


@pytest.mark.asyncio
async def test_nject_generic_factory(adapters):
    import typing
    from buvar import di

    FooType = typing.TypeVar("FooType", bound="Foo")

    class Foo:
        @classmethod
        async def adapt(cls: typing.Type[FooType]) -> FooType:
            return cls()

    class Bar(Foo):
        pass

    adapters.register(Foo.adapt)

    bar = await adapters.nject(Bar)
    assert isinstance(bar, Bar)


@pytest.mark.asyncio
async def test_nject_classmethod(adapters):
    class Foo:
        @classmethod
        async def adapt(cls) -> "Foo":
            return cls()

    adapters.register(Foo.adapt)

    foo = await adapters.nject(Foo)
    assert isinstance(foo, Foo)


@pytest.mark.asyncio
async def test_nject_class(adapters):
    class Foo:
        ...

    adapters.register(Foo)

    foo = await adapters.nject(Foo)
    assert isinstance(foo, Foo)


@pytest.mark.asyncio
async def test_nject_class_base(adapters):
    class Foo:
        ...

    class Bar(Foo):
        ...

    adapters.register(Bar)

    foo = await adapters.nject(Foo)
    assert isinstance(foo, Foo)


@pytest.mark.asyncio
async def test_nject_func(adapters):
    class Foo:
        ...

    async def adapt_foo() -> Foo:
        return Foo()

    adapters.register(adapt_foo)

    foo = await adapters.nject(Foo)
    assert isinstance(foo, Foo)


@pytest.mark.asyncio
async def test_nject_optional(adapters):
    import typing

    class Bar:
        ...

    class Foo:
        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs

    async def adapt_foo(bar: typing.Optional[Bar] = None) -> Foo:
        return Foo(bar=bar)

    adapters.register(adapt_foo)

    foo = await adapters.nject(Foo)
    assert foo.kwargs["bar"] is None


@pytest.mark.benchmark(group="nject")
def test_di_nject(event_loop, benchmark, adapters):
    from buvar import context

    class Foo(dict):
        def __init__(self, name=None):
            super().__init__(name=name, foo=True)

    class Bar(dict):
        def __init__(self, bar: Foo):
            super().__init__(foo=bar, bar=True)

    class Baz(dict):
        def __init__(self, **kwargs):
            super().__init__(baz=True, **kwargs)

    class Bim:
        pass

    class Bum(dict):
        def __init__(self):
            super().__init__(bum=True)

    async def baz_adapter(
        bar: Bar, bam: Foo = 1, bim: Bim = "default", *, bum: Bum, foo: Foo = None
    ) -> Baz:
        return Baz(foo=foo, bam=bam, bim=bim, bar=bar, bum=bum)

    adapters.register(Bar, baz_adapter)

    async def test():
        context.add(Foo())
        context.add(Foo(name="bar"), name="bar")

        baz, bum = await adapters.nject(Baz, Bum, bum=Bum())
        assert isinstance(baz, Baz)
        assert baz == {
            "baz": True,
            "bam": {"foo": True, "name": None},
            "foo": {"foo": True, "name": None},
            "bim": "default",
            "bum": {"bum": True},
            "bar": {"bar": True, "foo": {"foo": True, "name": "bar"}},
        }
        assert bum == {"bum": True}

    def bench():
        event_loop.run_until_complete(test())

    benchmark(bench)


@pytest.mark.benchmark(group="nject_2")
def test_nject_2(event_loop, benchmark, adapters):
    class Bar:
        pass

    class Foo:
        def __init__(self, bar: Bar = None):
            self.bar = bar

        @classmethod
        async def adapt_classmethod(cls, baz: str) -> "Foo":
            return Foo()

    async def adapt(bar: Bar) -> Foo:
        foo = Foo(bar)
        return foo

    adapters.register(Foo.adapt_classmethod, adapt)

    async def test():
        foo = await adapters.nject(Foo, baz="baz")
        assert foo.bar is None

        bar = Bar()
        foo = await adapters.nject(Foo, bar=bar)
        assert foo.bar is bar

    def bench():
        event_loop.run_until_complete(test())

    benchmark(bench)


@pytest.mark.asyncio
async def test_abc_meta_derived(adapters):
    import abc

    class Foo(abc.ABC):
        ...

    class Bar(Foo):
        ...

    adapters.register(Bar)

    bar = await adapters.nject(Bar)
    assert isinstance(bar, Bar)


@pytest.mark.asyncio
async def test_adapter_string_return(adapters):
    class Foo:
        @classmethod
        def adapt(cls, str: str) -> "Foo":
            return cls()

    adapters.register(Foo)

    foo = await adapters.nject(Foo)
    assert isinstance(foo, Foo)


@pytest.mark.asyncio
async def test_nject_deep_dependency_by_arg(adapters):
    class Foo:
        ...

    class Bar:
        def __init__(self, foo):
            self.foo = foo

        @classmethod
        def adapt_foo(cls, foo: Foo) -> "Bar":
            return cls(foo=foo)

    class Baz:
        def __init__(self, bar):
            self.bar = bar

        @classmethod
        def adapt_bar(cls, bar: Bar) -> "Baz":
            return cls(bar=bar)

    adapters.register(Bar.adapt_foo, Baz.adapt_bar)

    baz = await adapters.nject(Baz, foo=Foo())
    assert isinstance(baz, Baz)



"""A simple component architecture.

must have:
- add components by their type and a name
- add components by someother discriminator
- lookup components by their type/discriminator and name
- find components just by their type


nice to have:
- adaption resp. dependency injection of components
"""
import itertools
import typing as t
from buvar.components import ComponentLookupError

missing = object()


T = t.TypeVar("T")


class Components:
    """A component registry holds certain items identified by a
    namespace discriminator."""

    def __init__(self, *stack):
        self.stack = list(stack) or [{}]  # always at least one map
        self.namespaces = self.stack[0]

    def add(self, item, namespace=None, *, name=None):
        if isinstance(item, type):
            raise ValueError("A component should be an instance.")

        if namespace is None:
            namespace = type(item)

        if isinstance(namespace, type):
            mro = namespace.__mro__
        else:
            mro = [namespace]

        # we add the whole MRO to have a fast lookup later
        for namespace in mro:
            if namespace in self.namespaces:
                self.namespaces[namespace][name] = item
            else:
                space = {name: item}
                self.namespaces[namespace] = space
        return item

    def push(self, namespaces=None, *stack):
        """Push a whole stack on top of the actual components or just a new
        bare namespaces one.
        """
        if namespaces is None:
            namespaces = {}
        components = self.__class__(namespaces, *itertools.chain(stack, self.stack))
        return components

    def pop(self):
        return self.__class__(*self.stack[1:])

    def find(self, namespace):
        merged = {}
        for namespaces in self.stack[::-1]:
            try:
                space = namespaces[namespace]
                merged.update(space)
            except KeyError:
                pass
        return merged

    def get(self, namespace: t.Type[T], *, name=None, default=missing) -> T:
        for namespaces in self.stack:
            try:
                item = namespaces[namespace][name]
                return item
            except KeyError:
                pass

        if default is missing:
            raise ComponentLookupError("Component not found", namespace, name, default)
        return default



"""Very simple plugin architecture for asyncio.

`Loader` calls prepare coroutines, which yield tasks, the `Loader` collects.
These tasks are then run until complete.

A py:obj:`buvar.components.Components` context is stacked in way, that plugins
share the same context, while tasks don't, but may access the plugin context.

    >>> loop = asyncio.get_event_loop()
    >>> state = {}
    >>> async def prepare(load: Loader):
    ...     async def some_task():
    ...         state['task'] = True
    ...         return "foo"
    ...     yield some_task()

    >>> mystage = Stage()
    >>> mystage.load(prepare)
    >>> mystage.run()
    ['foo']
    >>> assert state == {'task': True}

    >>> state = {}
    >>> stage(prepare)
    ['foo']
    >>> assert state == {'task': True}
"""

# XXX FIXME doctest sometime shows log messages
import asyncio
import collections.abc
import inspect
import itertools
import sys
import types
import typing as t
import signal

import structlog

from . import context, util

PLUGIN_FUNCTION_NAME = "prepare"


sl = structlog.get_logger()


class Cancel(asyncio.Event):
    pass


class Teardown:
    """A collection of teardown tasks."""

    def __init__(self):
        self.tasks = []

    def add(self, task):
        self.tasks.append(task)

    def __iter__(self):
        """We teardown in the reverse order of registration."""
        return reversed(self.tasks)

    async def wait(self):
        # XXX TODO distinct between corotines, task, and normal functions
        await asyncio.gather(*self)


class Loader:
    """Load plugins and collect tasks."""

    def __init__(self):
        self._tasks = {}

    @property
    def tasks(self):
        return iter(itertools.chain(*self._tasks.values()))

    async def __call__(self, *plugins):
        """Hook the plugin from another plugin.

        The plugin may return an awaitable, an iterable of awaitables or an
        asyncgenerator of awaitables.

        The plugin function may use annotated arguments, which are provided by
        the actual stage context.

        :param plugins: the plugin to load
        :type plugins: list of callables
        """
        for plugin in plugins:
            plugin = resolve_plugin_func(plugin, caller=sys._getframe(1))
            if plugin not in self._tasks:
                sl.info("Plugin", plugin=util.fqdn(plugin))
                # mark plugin as loaded for recursive circular stuff
                self._tasks[plugin] = []
                # load plugin
                args = collect_plugin_args(plugin)
                try:
                    result = plugin(**args)
                except TypeError as ex:
                    raise TypeError(*ex.args, plugin.__module__, plugin.__name__)
                callables = [fun async for fun in generate_async_result(result)]
                self._tasks[plugin].extend(callables)


class Signals:
    handlers = (
        (signal.SIGINT, "handle_int"),
        (signal.SIGTERM, "handle_term"),
        (signal.SIGHUP, "handle_hup"),
        (signal.SIGQUIT, "handle_quit"),
        (signal.SIGABRT, "handle_abort"),
        (signal.SIGWINCH, "handle_winch"),
        (signal.SIGUSR1, "handle_usr1"),
        (signal.SIGUSR2, "handle_usr1"),
    )

    def __init__(self, stage):
        self.stage = stage
        self.add_signal_handlers()

    def add_signal_handlers(self):
        loop = self.stage.loop
        for signum, name in self.handlers:
            impl = getattr(self, name, None)
            if callable(impl):
                loop.add_signal_handler(signum, impl)

    # we really leave it open to the user how to handle signals
    # at least SIGINT makes sense
    def handle_int(self):
        self.stage.cancel.set()

    # def handle_term(self):
    #     ...

    # def handle_hup(self):
    #     ...

    # def handle_quit(self):
    #     ...

    # def handle_abort(self):
    #     ...

    # def handle_winch(self):
    #     ...

    # def handle_usr1(self):
    #     ...

    # def handle_usr2(self):
    #     ...


class Stage:
    """Stage manages the context stack while running each phase of the machinery.

    The stack typically has four layers:

    1. the default import time context

    2. the components context passed to the stage

    3. stage management context, e.g. Cancel, Teardown, Loader

    4. the shared plugin preparation context
    """

    def __init__(self, components=None, loop=None, signals: t.Type[Signals] = None):
        self.loop = loop or asyncio.get_event_loop()
        self.context = (
            context.current_context()
            .push(*(components.stack if components else ()))
            .push()
        )
        # provide basic components
        self.cancel = self.context.add(Cancel())
        self.teardown = self.context.add(Teardown())
        self.loader = self.context.add(Loader())
        self.signals = self.context.add((signals or Signals)(self))

        self.context = self.context.push()

    def load(self, *plugins):
        sl.info("Loading plugins")

        @context.run(self.context)
        def _load():
            self.loop.run_until_complete(self.loader(*plugins))

        _load()

    def run_tasks(self):
        sl.info("Running tasks", tasks=self.loader.tasks)

        @context.run(self.context)
        def _run_tasks():
            return self.loop.run_until_complete(
                run(tasks=self.loader.tasks, evt_cancel=self.cancel)
            )

        return _run_tasks()

    def run_teardown(self):
        sl.info("Teardown", tasks=self.teardown.tasks)
        self.loop.run_until_complete(self.teardown.wait())

    def run(self, *plugins):
        """Start the asyncio process by bootstrapping the root plugins.

        We have a three phase setup:

        1. run plugin hooks, to register arbitrary stuff and mainly gather a set of
        asyncio tasks for the second phase

        2. run all registered tasks together until complete

        3. teardown
        """

        try:
            # stage 1: bootstrap plugins
            self.load(*plugins)

            # stage 2: run main task and collect teardown tasks
            return self.run_tasks()
        finally:
            # stage 3: teardown
            self.run_teardown()


def stage(*plugins, components=None, loop=None, signals: t.Type[Signals] = None):
    if loop is None:
        loop = asyncio.get_event_loop()

    stage = Stage(components=components, loop=loop, signals=signals)
    return stage.run(*plugins)


async def run(tasks, *, evt_cancel=None):
    if evt_cancel is None:
        evt_cancel = context.add(Cancel())

    # we automatically elevate context stack for tasks
    factory = context.set_task_factory()
    try:
        fut_tasks = asyncio.gather(*map(asyncio.create_task, tasks))

        # stop staging if we finish in any way
        fut_tasks.add_done_callback(lambda _: evt_cancel.set())

        # wait for exit
        await evt_cancel.wait()

        if not fut_tasks.done():
            # we were explicitelly stopped by cancel event
            fut_tasks.cancel()
            try:
                await fut_tasks
            except asyncio.CancelledError:
                # silence our shutdown
                pass
        else:
            return fut_tasks.result()
    finally:
        factory.reset()


def collect_plugin_args(plugin):
    hints = t.get_type_hints(plugin)
    args = {name: context.get(cls) for name, cls in hints.items()}
    return args


def resolve_plugin_func(
    plugin,
    *,
    function_name: str = PLUGIN_FUNCTION_NAME,
    caller: t.Union[types.FrameType, int] = 0,
) -> t.Callable:
    plugin = util.resolve_dotted_name(
        plugin, caller=(caller + 1) if isinstance(caller, int) else caller
    )
    if inspect.ismodule(plugin):
        # apply default name
        plugin = getattr(plugin, function_name)

    if not (inspect.iscoroutinefunction(plugin) or inspect.isasyncgenfunction(plugin)):
        raise ValueError(f"{plugin} must a coroutine or an async generator.")

    return plugin


async def generate_async_result(fun_result):
    """Create an async generator out of any result."""
    if inspect.isasyncgen(fun_result):
        async for item in fun_result:
            yield item
    else:
        result = await fun_result
        if isinstance(result, collections.abc.Iterable):
            for item in result:
                yield item
        elif result is not None:
            yield result



from buvar import components, context
from .exc import missing, ResolveError


def _get_name_or_default(cmps, target, name=None):
    # find in components
    if name is not None:
        try:
            component = cmps.get(target, name=name)
            return component
        except components.ComponentLookupError:
            pass

    component = cmps.get(target, name=None)
    return component


class AdaptersImpl:
    async def nject(self, *targets, **dependencies):
        """Resolve all dependencies and return the created component."""
        # create components
        cmps = components.Components()

        # add default unnamed dependencies
        # every non-default argument of the same type gets its value
        # XXX is this good?
        for dep in dependencies.values():
            cmps.add(dep)

        # add current context
        current_context = context.current_context()
        stack = current_context.stack if current_context else []
        cmps = cmps.push(*stack)

        # add default named dependencies
        cmps = cmps.push()
        for name, dep in dependencies.items():
            cmps.add(dep, name=name)

        # find the proper components to instantiate that class
        injected = [
            await self.resolve_adapter(cmps, target, name=name)
            for name, target in ((None, target) for target in targets)
        ]
        if len(targets) == 1:
            return injected[0]
        return injected

    async def resolve_adapter(self, cmps, target, *, name=None, default=missing):
        # find in components
        try:
            component = _get_name_or_default(cmps, target, name)
            return component
        except components.ComponentLookupError:
            pass

        resolve_errors = []

        for adapter in self.lookup(target):
            try:
                adapter_args = {
                    param.name: await self.resolve_adapter(
                        cmps,
                        param.annotation,
                        name=param.name,
                        default=param.default,
                    )
                    for param in adapter.parameters.values()
                }
            except ResolveError as ex:
                # try next adapter
                resolve_errors.append(ex)
            else:
                component = await adapter.create(target, **adapter_args)
                # we do not use the name
                cmps.add(component)
                return component

        if default is not missing:
            return default

        if resolve_errors:
            raise ResolveError("No adapter dependencies found", target, resolve_errors)
        raise ResolveError("No possible adapter found", target, [])



import abc
import contextlib
import os
import queue
import signal
import socket
import typing as t
import structlog

import uritools

from buvar import plugin
from buvar.components import Components

sl = structlog.get_logger()
URI = uritools.SplitResult

SocketArgs = t.Optional[t.Tuple[t.Any, ...]]


class Socket(socket.socket, metaclass=abc.ABCMeta):
    __impls__ = set()

    def __new__(cls, ref: str) -> "Socket":
        parts: URI = uritools.urisplit(ref)
        for impl_cls in cls.__impls__:
            uri_socket_args = impl_cls.create(parts)
            if uri_socket_args:
                inst_uri, *socket_args = uri_socket_args
                inst = super().__new__(impl_cls, *socket_args)
                inst.uri = inst_uri
                inst.args = socket_args
                return inst

        raise TypeError(f"No socket implementation for: {ref}", ref)

    def __init__(self, uri):
        super().__init__(*self.args)

    def __init_subclass__(cls) -> None:
        cls.__impls__.add(cls)

    @abc.abstractmethod
    def bind(cls):
        ...

    @abc.abstractclassmethod
    def create(cls, uri: URI) -> SocketArgs:
        ...

    def __hash__(self):
        return hash(self.uri)

    def __eq__(self, other):
        return isinstance(other, Socket) and self.uri == other.uri

    def __str__(self):
        return uritools.uriunsplit(self.uri)


class TCPSocket(Socket):
    @classmethod
    def create(cls, uri: URI) -> SocketArgs:
        if uri.scheme == "tcp":
            if uri.gethost() == "":
                uri = uri.transform(f"tcp://0.0.0.0:{uri.getport()}")

            return uri, socket.AF_INET, socket.SOCK_STREAM

    def bind(self):
        socket.socket.bind(self, (str(self.uri.gethost()), self.uri.getport()))


class UnixSocket(Socket):
    @classmethod
    def create(cls, uri: URI) -> SocketArgs:
        if uri.scheme == "unix":
            # resolve abs path
            ...
            return uri, socket.AF_UNIX, socket.SOCK_STREAM

    def bind(self):
        super().bind(self.uri.getpath())


class Sockets(dict):
    def __init__(self, *sockets: str):
        super().__init__((str(s), s) for s in map(Socket, sockets))

    @contextlib.contextmanager
    def bind(self):
        pid = os.getpid()
        try:
            for s in self.values():
                s.bind()
            yield self
        finally:
            if pid == os.getpid():
                for s in self.values():
                    s.close()


class Fork:
    number: int = 0
    ppid: int
    pid: int
    is_child: bool

    def __init__(self, number: int):
        self.number = number
        self.children = set()
        self.queue = queue.Queue()
        self.pid = os.getpid()
        self.ppid = os.getppid()
        self.is_child = False

    def run(
        self, func: t.Callable, *args: t.Any, **kv: t.Any
    ) -> t.Optional[t.List[t.Any]]:
        forks = len(os.sched_getaffinity(0)) if not self.number else self.number

        if forks == 1:
            sl.debug("Skip forking", forks=forks, parent=os.getpid())
            return func(*args, **kv)

        sl.debug("Forking", forks=forks, parent=os.getpid())
        for i in range(forks):
            child = os.fork()
            if child:
                sl.debug("Child", child=child)
                self.children.add(child)
            else:
                # override for child
                self.pid = os.getpid()
                self.ppid = os.getppid()
                self.is_child = True
                sl.debug("Run", child=self.pid, parent=self.ppid, func=func)
                result = func(*args, **kv)
                sl.debug("Stopped", child=self.pid, parent=self.ppid, func=func)
                self.queue.put(result)
                # stop child from iterating the rest
                return

        sl.debug("Waiting for children to exit", children=self.children)
        self.wait_for_children()
        result = []
        while True:
            try:
                result.append(self.queue.get(False))
            except queue.Empty:
                break
        return result

    def _signal_children(self, signum, frame):
        for child in self.children:
            os.kill(child, signum)

    def wait_for_children(self):
        # forward signal to children
        for signum in (
            signal.SIGINT,
            signal.SIGTERM,
            signal.SIGHUP,
            signal.SIGQUIT,
            signal.SIGABRT,
            signal.SIGWINCH,
            signal.SIGUSR1,
            signal.SIGUSR2,
        ):
            signal.signal(signum, self._signal_children)
        # just wait for them
        pgid = os.getpgid(self.pid)
        os.waitid(os.P_PGID, pgid, os.WEXITED)


def stage(
    *plugins,
    components=None,
    loop=None,
    forks: int = 0,
    sockets: t.Optional[t.Sequence[str]] = None,
):
    if components is None:
        components = Components()

    f = components.add(Fork(forks))
    with Sockets(*sockets).bind() as s:
        # register sockets
        for name, s in s.items():
            components.add(s, name=name)

        result = f.run(plugin.stage, *plugins, components=components, loop=loop)
        return result



import pytest


def foo():
    ...


def test_resolve_dotted_name():
    from buvar import util

    fun = util.resolve_dotted_name("dotted:name")

    assert fun() == "foobar"


def test_resolve_dotted_name_no_module():
    from buvar import util

    with pytest.raises(ModuleNotFoundError):
        util.resolve_dotted_name("nomodule:name")


def test_resolve_dotted_name_wrong_name():
    from buvar import util

    with pytest.raises(ValueError):
        util.resolve_dotted_name("nomodule:na:me")


def test_resolve_dotted_name_relative():
    from buvar import util

    func = util.resolve_dotted_name(".test_util:foo")

    assert func is foo



"""Dependency injection.

Register a class, function or method to adapt arguments into something else.

>>> class Bar:
...     pass

>>> class Foo:
...     def __init__(self, bar: Bar):
...         self.bar = bar
...     @classmethod
...     def adapt(cls, bar: Bar) -> 'Foo':
...         return cls(bar)

>>> def adapt() -> Bar:
...     return Bar()

>>> adapters = Adapters()
>>> adapters.register(Bar)
>>> adapters.register(Foo.adapt)
>>> adapters.register(adapt)

>>> import asyncio
>>> loop = asyncio.get_event_loop()
>>> foo = loop.run_until_complete(adapters.nject(Foo))
>>> assert isinstance(foo, Foo)

"""

import abc
import contextvars as cv
import functools as ft
import inspect
import itertools as it
import sys
import typing as t

import typing_inspect as ti

from buvar import util

from .exc import ResolveError, missing

try:
    # gains over 100% speed up
    from . import c_di as _impl

except ImportError:
    from . import py_di as _impl


PY_39 = sys.version_info >= (3, 9)


class AdapterError(Exception):
    pass


class Adapters(dict, _impl.AdaptersImpl):
    def __init__(self):
        super().__init__()
        self.context = cv.copy_context()

    def __hash__(self):
        # since we want to cache our lookup, we need have to be hashable
        return object.__hash__(self)

    def register(self, *implementations, frame=None, **kwargs):
        frame = frame or sys._getframe(1)
        for implementation in implementations:
            adapter = Adapter(implementation, frame=frame)
            adapter.register(self, **kwargs)
        # self.lookup.cache_clear()

    # @ft.lru_cache(maxsize=None)
    def lookup(self, tp):
        errors = {}

        def collect():
            for adapter_cls in reversed(Adapter.classes):
                try:
                    yield from adapter_cls.lookup(self, tp)
                except Exception as ex:
                    errors[adapter_cls] = ex

        return set(collect())


class AdapterMeta(abc.ABCMeta):
    classes = []

    def __call__(cls, implementation, *, frame=None):
        frame = frame or sys._getframe(1)
        errors = {}
        # polymorphic instantiation: trial and error from specific to generic
        for adapter_cls in reversed(cls.classes):
            try:
                adapter = type.__call__(adapter_cls, implementation, frame=frame)
                return adapter
            except Exception as ex:
                errors[adapter_cls] = ex
        if errors:
            raise AdapterError("All adaptations failed", errors)
        raise AdapterError("No adapters")


class Adapter(metaclass=AdapterMeta):
    def __init__(self, implementation, **_):
        self.implementation = implementation

    def __init_subclass__(cls):
        if not inspect.isabstract(cls):
            cls.classes.append(cls)

    @abc.abstractmethod
    def register(self, registry: t.Dict):
        """Register this adapter into the registry."""

    @abc.abstractclassmethod
    def lookup(cls, registry: t.Dict, tp) -> t.Iterator:
        """Lookup an adapter based on its return type."""

    if sys.version_info >= (3, 8):

        async def create(self, target, *args, **kwargs):
            call = self.implementation(*args, **kwargs)
            return (
                await call if inspect.iscoroutinefunction(self.implementation) else call
            )

    else:

        async def create(self, target, *args, **kwargs):
            call = self.implementation(*args, **kwargs)
            if isinstance(self.implementation, ft.partial):
                return (
                    await call
                    if inspect.iscoroutinefunction(self.implementation.func)
                    else call
                )
            else:
                return (
                    await call
                    if inspect.iscoroutinefunction(self.implementation)
                    else call
                )


def evaluate(tp: t.Union[str, t.TypeVar, t.Type, t.ForwardRef], *, frame=None):
    if isinstance(tp, str):
        tp = t.ForwardRef(tp)

    if ti.is_typevar(tp):
        tp = ti.get_bound(tp)

    # TODO python versions
    return t._eval_type(
        tp,
        frame.f_globals if frame else None,
        frame.f_locals if frame else None,
    )


def evaluated_signature(func: t.Callable, frame=None):
    """Adjust annotations."""
    signature = inspect.Signature.from_callable(func)
    signature = signature.replace(
        parameters=[
            p.replace(
                annotation=evaluate(p.annotation, frame=frame),
                default=p.default
                # we change empty to missing for later convenience
                if p.default is not inspect.Signature.empty
                else missing,
            )
            for p in signature.parameters.values()
        ],
        return_annotation=evaluate(signature.return_annotation, frame=frame),
    )
    return signature


class BaseMatrix:
    def __call__(self, tp):
        if inspect.isclass(tp):
            return self.iter_mro(tp)
        if ti.is_optional_type(tp):
            return self.iter_optional(tp)
        if ti.is_tuple_type(tp):
            return self.iter_generic(tp)
        if ti.is_union_type(tp):
            return self.iter_generic(tp)
        if ti.is_generic_type(tp):
            return self.iter_generic(tp)

    def iter_mro(self, tp):
        mro = inspect.getmro(tp)
        # skip object
        yield from iter(mro[:-1])

    def iter_optional(self, tp):
        arg, *_ = ti.get_args(tp, evaluate=True)
        for base in self(arg):
            yield t.Optional[base]

    def iter_generic(self, tp):
        if hasattr(tp, "copy_with"):
            copy_with = tp.copy_with
        else:
            # INFO: using dict[str, ...] is also generic, but dict provides no copy_with
            copy_with = ft.partial(tp.__class__, tp.__origin__)
        args = ti.get_args(tp, evaluate=True)
        yield from (copy_with(params) for params in it.product(*map(self, args)))


base_matrix = BaseMatrix()


class CallableAdapter(Adapter, inspect.Signature):
    def __init__(self, implementation, *, frame=None):
        if not callable(implementation):
            raise AdapterError("Implementation not callable", implementation)
        super().__init__(implementation)
        self.frame = frame
        signature = evaluated_signature(self.implementation, frame=frame)
        inspect.Signature.__init__(
            self,
            list(signature.parameters.values()),
            return_annotation=signature.return_annotation,
        )

    def __hash__(self):
        # INFO:we hash the signature and the implementation, because signature
        # may be the same for different implememtations
        return hash(
            (super(inspect.Signature, self).__hash__(), hash(self.implementation))
        )

    def __repr__(self):
        return f"<{self.__class__.__name__}[{self.implementation}] {self.parameters}"

    @util.cached
    def return_type(self):
        rt = self.return_annotation
        if rt is not inspect.Signature.empty and rt is not None:
            return rt
        if inspect.isclass(self.implementation):
            return self.implementation
        raise AdapterError(
            "Adapter implementation has no return type", self.implementation
        )


class GenericAdapter(CallableAdapter):
    def __init__(self, implementation, *, frame=None):
        super().__init__(implementation, frame=frame)

    @classmethod
    def registry(cls, registry: t.Dict):
        registry = registry.setdefault(cls, {})
        return registry

    def register(self, registry: t.Dict, replace: bool = False):
        registry = self.registry(registry)
        for base in base_matrix(self.return_type):
            if replace:
                registry[base] = {self}
            else:
                registry.setdefault(base, set()).add(self)

    @classmethod
    def lookup(cls, registry: t.Dict, tp):
        registry = cls.registry(registry)
        adapters = registry.get(tp)
        if adapters:
            yield from adapters


class MethodAdapter(GenericAdapter):
    def __init__(self, implementation, *, frame=None):
        super().__init__(implementation, frame=frame)
        if not inspect.ismethod(implementation):
            raise AdapterError("Implementation is not a method", implementation)


class ClassmethodAdapter(MethodAdapter):
    def __init__(self, implementation, *, frame=None):
        super().__init__(implementation, frame=frame)
        if not inspect.isclass(implementation.__self__):
            raise AdapterError("Implementation is not a classmethod", implementation)
        self.cls = implementation.__self__

    @classmethod
    def lookup(cls, registry: t.Dict, tp):
        """
        Foo <- Bar(Foo) <- Baz(Bar)
          |
          +-- adapt(cls) -> "Foo"

        An implementation returning parent is registered.

        We want to have an instance of `Bar` or `Baz`.
        If we do not find an adapter directly, we try to find an adapter, which
        returns a subclass of `Bar` or `Baz` and whose class is also this
        subclass
        """
        yield from super().lookup(registry, tp)
        local_registry = cls.registry(registry)
        # try inheritance
        for base in inspect.getmro(tp)[:-1]:
            for adapter in local_registry.get(base, ()):
                # found an adater which returns a base
                if issubclass(base, adapter.cls):
                    # found an adapter, whose class matches what he returns
                    # we prebind our implementation
                    bound_adapter = adapter.replace(
                        # TODO XXX FIXME partial only works in >=3.8
                        # create is not recognizing coroutine
                        ft.partial(adapter.implementation.__func__, tp)
                    )
                    # cache
                    local_registry.setdefault(tp, set()).add(bound_adapter)
                    yield bound_adapter

    def replace(self, implementation):
        adapter = type(self)(implementation, frame=self.frame)
        return adapter


# our default adapters
buvar_adapters = cv.ContextVar(__name__)
buvar_adapters.set(Adapters())


def register(*impls, **kwargs):
    adapters = buvar_adapters.get()
    adapters.register(*impls, frame=sys._getframe(1), **kwargs)


async def nject(*targets, **dependencies):
    adapters = buvar_adapters.get()
    return await adapters.nject(*targets, **dependencies)



import pytest


async def prepare():
    from buvar import context

    context.add("foobar")


@pytest.mark.asyncio
@pytest.mark.buvar_plugins("tests.test_testing")
async def test_wrapped_stage_context():
    from buvar import context, plugin

    assert context.get(str) == "foobar"
    assert context.get(plugin.Cancel)


@pytest.mark.asyncio
@pytest.mark.buvar_plugins()
async def test_wrapped_stage_context_load(buvar_load):
    from buvar import context, plugin

    await buvar_load(prepare)
    assert context.get(str) == "foobar"
    assert context.get(plugin.Cancel)



"""Provide a component registry as contextvar."""

import asyncio
import contextlib
import contextvars
import functools
import sys

from . import components

buvar_context: contextvars.ContextVar[components.Components] = contextvars.ContextVar(
    __name__
)

# we provide a globally available context
buvar_context.set(components.Components())


class StackingTaskFactory:
    def __init__(self, *, parent_factory=None):
        self.parent_factory = parent_factory

    if sys.version_info < (3, 11):

        def __call__(self, loop, coro, context=None):
            component_context = current_context().push()
            token = buvar_context.set(component_context)
            # with child():
            task = (
                self.parent_factory
                if self.parent_factory is not None
                else asyncio.tasks.Task
            )(loop=loop, coro=coro)
            try:
                return task
            finally:
                buvar_context.reset(token)
    else:
        # INFO: Task() accepts context
        def __call__(self, loop, coro, context=None):
            component_context = current_context().push()
            token = buvar_context.set(component_context)
            # with child():
            task = (
                self.parent_factory
                if self.parent_factory is not None
                else asyncio.tasks.Task
            )(loop=loop, coro=coro, context=context)
            try:
                return task
            finally:
                buvar_context.reset(token)

    @classmethod
    def set(cls, *, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()

        factory = cls(parent_factory=loop.get_task_factory())
        loop.set_task_factory(factory)
        return factory

    def reset(self, *, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()

        loop.set_task_factory(self.parent_factory)


set_task_factory = StackingTaskFactory.set


def current_context():
    return buvar_context.get()


def add(*args, **kwargs):
    context = current_context()
    return context.add(*args, **kwargs)


def get(*args, **kwargs):
    context = current_context()
    return context.get(*args, **kwargs)


def find(*args, **kwargs):
    context = current_context()
    return context.find(*args, **kwargs)


def push(*stack):
    context = current_context().push(*stack)
    return context


def pop():
    context = current_context().pop()
    return context


# https://www.python.org/dev/peps/pep-0568/
# https://stackoverflow.com/questions/53611690/how-do-i-write-consistent-stateful-context-managers
@contextlib.contextmanager
def child(*stack):
    context = push(*stack)
    token = buvar_context.set(context)
    try:
        yield
    finally:
        buvar_context.reset(token)


def run(context: components.Components):
    """Run the decorated function within the provided context."""

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            def _child_context():
                buvar_context.set(context)
                return func(*args, **kwargs)

            ctx = contextvars.copy_context()
            return ctx.run(_child_context)

        return wrapper

    return decorator



import pytest


def test_tcp_socket():
    import socket
    from buvar import fork

    s = fork.Socket("tcp://:12345")
    assert s.family == socket.AF_INET
    assert s.uri.getport() == 12345
    assert str(s.uri.gethost()) == "0.0.0.0"


def test_unix_socket():
    import socket
    from buvar import fork

    s = fork.Socket("unix:///tmp/foo.sock")
    assert s.family == socket.AF_UNIX
    assert s.uri.getpath() == "/tmp/foo.sock"


def test_tcp_sockets():
    from buvar import fork

    s = fork.Sockets(
        "tcp://:12345",
        "tcp://:12345",
        "unix:///tmp/foo.sock",
    )
    assert s == {
        "tcp://0.0.0.0:12345": fork.Socket("tcp://:12345"),
        "unix:///tmp/foo.sock": fork.Socket("unix:///tmp/foo.sock"),
    }


def test_socket_not_implemented():
    from buvar import fork

    with pytest.raises(TypeError):
        fork.Socket("http://:12345")


def test_bind_tcp_socket(Anything):
    from buvar import fork

    s = fork.Socket("tcp://:0")
    s.bind()
    try:
        assert s.getsockname() == ("0.0.0.0", Anything)
    finally:
        s.close()


def test_bind_sockets(Anything):
    from buvar import fork

    with fork.Sockets("tcp://:0").bind() as sockets:
        s = next(iter(sockets.values()))
        assert s.getsockname() == ("0.0.0.0", Anything)



import pytest


@pytest.fixture
def adapters():
    from buvar import di

    return di.Adapters()


@pytest.fixture
def Foo():
    import dataclasses as dc

    @dc.dataclass
    class Foo:
        bim: str
        bam: int = 123
        baz: bool = False

        @classmethod
        def adapt_children(cls) -> "Foo":
            return cls(bim=cls.__name__)

    return Foo


@pytest.fixture
def Bar(Foo):
    import dataclasses as dc
    import typing as t

    @dc.dataclass
    class Bar(Foo):
        @classmethod
        def adapt(cls, bim: str) -> t.Optional["Bar"]:
            return Bar(bim="bam")

    return Bar


@pytest.mark.asyncio
async def test_create_adapter_callable(adapters, Foo, Bar):
    import typing as t

    async def adapt_list_of_bar() -> t.List[Bar]:
        return [Bar(bim="foo")]

    adapters.register(adapt_list_of_bar)
    adapters.register(Bar)
    adapters.register(Bar.adapt)

    assert len(list(adapters.lookup(Foo))) == 1
    assert list(adapters.lookup(Bar))
    assert list(adapters.lookup(t.Optional[Foo]))
    assert list(adapters.lookup(t.List[Foo]))

    assert await adapters.nject(t.List[Foo])
    assert await adapters.nject(t.Optional[Bar], bim="foo")
    assert isinstance(await adapters.nject(Foo, bim="foo"), Foo)
    assert isinstance(await adapters.nject(Bar, bim="bar"), Bar)


@pytest.mark.benchmark(group="lookup")
def test_adapters_lookup_benchmark(adapters, Foo, Bar, benchmark):
    import typing as t

    async def adapt_list_of_bar() -> t.List[Bar]:
        return [Bar(bim="foo")]

    adapters.register(adapt_list_of_bar)
    adapters.register(Bar)
    adapters.register(Bar.adapt)

    def lookup():
        assert len(list(adapters.lookup(Foo))) == 1
        assert list(adapters.lookup(Bar))
        assert list(adapters.lookup(t.Optional[Foo]))
        assert list(adapters.lookup(t.List[Foo]))

    benchmark(lookup)


@pytest.mark.asyncio
async def test_create_adapter_inheritance(adapters):
    import typing as t

    class Baz:
        ...

    class Foo:
        def __init__(self, bim: str):
            self.bim = bim

        @classmethod
        def adapt_children(cls) -> "Foo":
            return cls(bim=cls)

        @classmethod
        def adapt_other(cls) -> Baz:
            baz = Baz()
            baz.adapter = cls
            return baz

    class Bar(Foo):
        ...

    adapters.register(Foo.adapt_children, Foo.adapt_other)

    bar = await adapters.nject(Bar)
    assert bar.bim == Bar
    bar = await adapters.nject(Bar)
    assert bar.bim == Bar

    baz = await adapters.nject(Baz)
    assert baz.adapter == Foo



import pytest


@pytest.mark.benchmark(group="get")
def test_components_get(benchmark):
    from buvar import components

    class Foo:
        pass

    class Bar(Foo):
        pass

    class Baz(Bar):
        pass

    c = components.Components()

    foo = Foo()
    baz = Baz()
    c.add(foo)
    c.add(baz)

    def test():
        assert c.get(Bar) == baz
        assert c.get(Foo) == baz
        assert c.get(Baz) == baz

    benchmark(test)


def test_components_get_basic():
    from buvar import components

    c = components.Components()

    c.add("foo", "bar")

    assert c.get("bar") == "foo"


def test_components_get_bytes_name():
    from buvar import components

    c = components.Components()

    c.add("foo", b"bar")

    assert c.get(b"bar") == "foo"


def test_components_no_instance():
    from buvar import components

    c = components.Components()

    class Foo:
        pass

    with pytest.raises(ValueError):
        c.add(Foo)


def test_components_deep_find():
    from buvar import components

    c = components.Components()
    c.add("foo")
    cc = c.push()
    cc.add("bar", name="bar")

    assert c.find(str) == {None: "foo"}
    assert cc.find(str) == {None: "foo", "bar": "bar"}


@pytest.fixture
def components():
    from buvar import components

    yield components


@pytest.mark.benchmark(group="find")
def test_components_find(benchmark, components):
    class Foo:
        pass

    class Bar(Foo):
        pass

    c1 = components.Components()
    c2 = c1.push()
    c3 = c2.push()
    c4 = c3.push()

    foo = c1.add(Foo())
    bar = c2.add(Bar(), name="bar")
    foobar = c3.add(Foo(), name="bar")
    barbar = c4.add(Bar(), name="bar")

    def find():
        assert c3.find(Foo) == {None: foo, "bar": foobar}
        assert c3.find(Bar) == {"bar": bar}
        assert c4.find(Bar) == {"bar": barbar}

        assert c3.get(Bar, name="bar") == bar
        assert c4.get(Bar, name="bar") == barbar

    benchmark(find)


def test_componentsn_find_str(components):
    c = components.Components()
    c.add("foo", "foo")

    space = c.find("foo")
    assert space == {None: "foo"}


@pytest.mark.benchmark(group="push pop")
def test_components_push_pop(benchmark, components):
    c = components.Components()
    c.add("foo")

    def pushpop():
        cc = c.push()
        cc.add("bar", name="bar")
        assert c.find(str) == cc.pop().find(str)

    benchmark(pushpop)

    assert c.find(str) == {None: "foo"}


def test_components_add_name_any(components):
    c = components.Components()

    c.add("foo", name=("bar", 123))

    foo = c.get(str, name=("bar", 123))

    assert foo == "foo"



import itertools

from setuptools import dist, find_packages, setup
from setuptools.command.install import install

from build import build


class BinaryDistribution(dist.Distribution):
    def is_pure(self):
        return False

    def has_ext_modules(self):
        return True


class InstallPlatlib(install):
    def finalize_options(self):
        install.finalize_options(self)
        if self.distribution.has_ext_modules():
            self.install_lib = self.install_platlib


with open("README.rst") as f:
    description = f.read()


install_requires = [
    "attrs",
    "cattrs",
    "structlog>=20.1.0",
    "toml>=0.10",
    "typing_inspect>=0.4.0",
    "cached_property",
    "uritools",
]
extras_require = {
    "tests": [
        "pytest>=4.6",
        "pytest-cov>=2.7",
        "pytest-asyncio>=0.11.0",
        "pytest-benchmark>=3.2.2",
        "mock>=3.0",
        "pytest-mock>=1.10",
        "pytest-watch>=4.2",
        "pytest-randomly>=3.1",
        "pytest-anything",
    ],
    "dev": [
        "commitizen",
        "pre-commit",
        "pdbpp",
        "ipython",
    ],
}
extras_require["all"] = list(itertools.chain(extras_require.values()))


entry_points = {"pytest11": ["buvar = buvar.testing"]}

setup_kwargs = {
    "name": "buvar",
    "version": "0.43.22",
    "description": "Asyncio plugins, components, dependency injection and configs",
    "long_description": description,
    "long_description_content_type": "text/x-rst",
    "author": "Oliver Berger",
    "author_email": "diefans@gmail.com",
    "url": "https://github.com/diefans/buvar",
    "package_dir": {"": "src"},
    "packages": find_packages("src"),
    "include_package_data": True,
    # "package_data": {"": ["*"]},
    "install_requires": install_requires,
    "extras_require": extras_require,
    "entry_points": entry_points,
    "python_requires": ">=3.7,<4.0",
    "classifiers": [
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Framework :: AsyncIO",
        "License :: OSI Approved :: Apache Software License",
        "License :: OSI Approved :: MIT License",
    ],
    "cmdclass": {"install": InstallPlatlib},
    "distclass": BinaryDistribution,
}
build(setup_kwargs)

setup(**setup_kwargs)
